/*

Copyright 1993, 1994, Cornell University

Cornell hereby grants permission to use, copy, modify, and distribute this program for any purpose 
and without fee, provided that these copyright and permission notices appear on all copies and 
supporting documentation, the name of Cornell not be used in advertising or publicity pertaining 
to distribution of the program without specific prior permission, notice be given in supporting 
documentation that copying and distribution is by permission of Cornell.  CORNELL MAKES NO 
REPRESENTATIONS OR WARRANTEES, EXPRESS OR IMPLIED.  By way of example, but not limitation, 
CORNELL MAKES NO REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR 
PURPOSE OR THAT THE USE OF THIS SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, 
TRADEMARKS, OR OTHER RIGHTS.  Cornell shall not be held liable for any liability with respect to 
any claim by the user or any other party arising from use of the program.

This material is partially based on work sponsored by the National Science Foundation under Cooperative 
Agreement No. NCR-9318337.  The government has certain rights in this material.

*/



#define MAXDATALEN 8096 
#define MINREFPKT   4

/*  monitor requests		*/

#define VERSION 	1			/* return the version */
#define WHO 	        2			/* return conference participants */
#define TERM 	        3			/* terminate the reflector        */
#define PARAM 	        4			/* return the paramater list      */
#define MAVEN           5                       /* return the MAVEN client list   */
#define UPTIME          6                       /* return the start time          */

typedef struct RefConPkt
{
   short msg_type;
   short msg_len;
   char  msg_data[MAXDATALEN];
} RefConPkt;
